/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2021  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QSpacerItem>
#include <QWidget>
#include <QCloseEvent>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QMimeData>
#include <QUrl>
#include <QDir>
#include <QColor>
#include <QFile>
#include <QFileDialog>
#include <QFont>
#include <cmath>
#include <assert.h>

#include "../../core/graphwidget.h"
#include "../../core/vector2.h"
#include "../../core/interpolation.h"
#include "../../core/messagebox.h"
#include "../../../core/xmlfile.h"

#include "entrywidget.h"
#include "types.h"

#include "lengthcompare.h"

// sort pnt by position
struct Position_Sort
{
	bool operator() (double a,double b)
	{
		return (a < b);
	}
};

TLengthCompare::TLengthCompare(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator1_hline;
	QFrame								*separator2_hline;
	QGridLayout							*gridLayout;
	QHBoxLayout							*button_hlayout;
	QLabel								*analysis_label;
	QLabel								*length_input_label;
	QLabel								*copyright_label;
	QSpacerItem							*button_hspacer;
	QSpacerItem							*button_vspacer;
	QFont								copyright_font;

	this->resize(901,801);
	this->setAcceptDrops(true);

	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	gridLayout = new QGridLayout(this);

	length_input_label = new QLabel(this);
	length_input_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	gridLayout->addWidget(length_input_label,0,0,1,1);
	
	d_load_button = new QPushButton(this);
	d_load_button->setAutoDefault(false);
	gridLayout->addWidget(d_load_button,1,0,1,1);
	
	d_save_button = new QPushButton(this);
	d_save_button->setAutoDefault(false);
	gridLayout->addWidget(d_save_button,2,0,1,1);
	
	d_export_button = new QPushButton(this);
	d_export_button->setAutoDefault(false);
	gridLayout->addWidget(d_export_button,3,0,1,1);
	
	button_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	gridLayout->addItem(button_vspacer,4,0,1,1);

	d_entry_a_widget = new TEntryWidget(this);
	gridLayout->addWidget(d_entry_a_widget,0,1,5,1);

	d_entry_b_widget = new TEntryWidget(this);
	gridLayout->addWidget(d_entry_b_widget,0,2,5,1);

	separator1_hline = new QFrame(this);
	separator1_hline->setFrameShape(QFrame::HLine);
	separator1_hline->setFrameShadow(QFrame::Sunken);
	gridLayout->addWidget(separator1_hline,5,0,1,3);

	analysis_label = new QLabel(this);
	analysis_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	gridLayout->addWidget(analysis_label,6,0,1,1);

	d_graph_data_widget = new TGraphWidget(this);
	d_graph_data_widget->setMinimumSize(QSize(100, 200));
	gridLayout->addWidget(d_graph_data_widget,6,1,1,2);
	
	d_graph_en_widget = new TGraphWidget(this);
	d_graph_en_widget->setMinimumSize(QSize(100, 200));
	gridLayout->addWidget(d_graph_en_widget,7,1,1,2);

	separator2_hline = new QFrame(this);
	separator2_hline->setFrameShape(QFrame::HLine);
	separator2_hline->setFrameShadow(QFrame::Sunken);
	gridLayout->addWidget(separator2_hline,8,0,1,3);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	d_close_button = new QPushButton(this);
	d_close_button->setAutoDefault(false);
	button_hlayout->addWidget(d_close_button);
	
	gridLayout->addLayout(button_hlayout,9,0,1,3);
	
	copyright_label = new QLabel(this);
	copyright_label->setFont(copyright_font);
	gridLayout->addWidget(copyright_label,10,0,1,3);


	this->setWindowTitle(QStringLiteral("Length Compare - Version 1.0"));
	
	length_input_label->setText(QStringLiteral("Length Data:"));
	analysis_label->setText(QStringLiteral("Analysis:"));
	d_load_button->setText(QStringLiteral("Load"));
	d_save_button->setText(QStringLiteral("Save"));
	d_export_button->setText(QStringLiteral("Export"));
	d_close_button->setText(QStringLiteral("Close"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2021 Select Calibration Incorporated.  www.selectcalibration.ca"));

	d_graph_data_widget->Title(QStringLiteral("Length Comparison"));
	d_graph_data_widget->VerticalLegend("Deviation");
	d_graph_data_widget->HorizontalLegend("Length");
	
	d_graph_en_widget->Title(QStringLiteral("Deviation Comparison"));
	d_graph_en_widget->VerticalLegend("En Score");
	d_graph_en_widget->HorizontalLegend("Length");
	
	d_save_button->setEnabled(false);
	d_export_button->setEnabled(false);
	
	d_message_box = new TMessageBox(this);
	
	connect(d_load_button,SIGNAL(clicked(void)),this,SLOT(Load(void)));
	connect(d_save_button,SIGNAL(clicked(void)),this,SLOT(Save(void)));
	connect(d_export_button,SIGNAL(clicked(void)),this,SLOT(Export(void)));
	connect(d_close_button,SIGNAL(clicked(void)),this,SLOT(close(void)));
	
	connect(d_entry_a_widget,SIGNAL(Data_Changed(void)),this,SLOT(Entry_Data_Changed(void)));
	connect(d_entry_b_widget,SIGNAL(Data_Changed(void)),this,SLOT(Entry_Data_Changed(void)));
	
//	this->Generate_Test_Data();
	this->Entry_Data_Changed();
	d_data_changed = false;
}

TLengthCompare::~TLengthCompare(void)
{
}

bool TLengthCompare::Save_Data(
	const QString						&file_name) const
{
	QString								text;
	std::vector<Type::TEntryItem>		input_a_data;
	std::vector<Type::TEntryItem>		input_b_data;
	std::vector<Type::TEntryItem>::const_iterator entry_iter;
	TXmlFile							xml_file;
	QDomElement							e0;
	QDomElement							e1;
	QDomElement							e2;

	try
	{
		xml_file.Set_File_Name(file_name);
		xml_file.Reset(QStringLiteral("Length_Compare_Data"),1);
		
		input_a_data = d_entry_a_widget->Entry_Items();
		
		// data set a
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Data_Set_A"));
		
		for(entry_iter = input_a_data.begin();entry_iter != input_a_data.end();++entry_iter)
		{
			e1 = xml_file.Create_Node(&e0,QStringLiteral("Entry"));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Nominal"));
			xml_file.Write_Text_Node(&e2,QString("%1").arg((*entry_iter).nominal,0,'f',5));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Actual"));
			xml_file.Write_Text_Node(&e2,QString("%1").arg((*entry_iter).actual,0,'f',5));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Uc"));
			xml_file.Write_Text_Node(&e2,QString("%1").arg((*entry_iter).uc,0,'f',5));
			
		}
		
		input_b_data = d_entry_b_widget->Entry_Items();
		
		// data set b
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Data_Set_B"));
		
		for(entry_iter = input_b_data.begin();entry_iter != input_b_data.end();++entry_iter)
		{
			e1 = xml_file.Create_Node(&e0,QStringLiteral("Entry"));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Nominal"));
			xml_file.Write_Text_Node(&e2,QString("%1").arg((*entry_iter).nominal,0,'f',5));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Actual"));
			xml_file.Write_Text_Node(&e2,QString("%1").arg((*entry_iter).actual,0,'f',5));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Uc"));
			xml_file.Write_Text_Node(&e2,QString("%1").arg((*entry_iter).uc,0,'f',5));
		}
		
		xml_file.Write_File();
	}
	catch(TXmlFileException exception)
	{
		d_last_error = QString("ERR  Save_Data: %1").arg(exception.ErrorText()).toLatin1();
		return false;
	}

	return true;
}

bool TLengthCompare::Export_Data(
	const QString						&file_name)
{
	QFile								file;
	std::vector<Type::TEntryItem>		input_a_data;
	std::vector<Type::TEntryItem>		input_b_data;
	std::vector<Type::TEntryItem>::const_iterator entry_iter;
	std::vector<TVector2>				dev_a_data;
	std::vector<TVector2>				dev_b_data;
	std::vector<TVector2>				uc_a_data;
	std::vector<TVector2>				uc_b_data;
	TVector2							pnt;
	TVector2							p1,p2;
	Position_Sort						position_sort;
	std::vector<double>::const_iterator	position_iter;
	QString								text;
	double								min_a_position(0);
	double								max_a_position(0);
	double								min_b_position(0);
	double								max_b_position(0);
	double								min_position(0);
	double								max_position(0);
	double								dev_a,dev_b;
	double								uc_a,uc_b;
	double								difference;
	double								uc_ab;
	double								en;
	double								en_mean;
	double								en_max;
	int									en_count;
	bool								init;
	static const double					MAX_EXTRAPOLATION_LENGTH(10.0);
	static const double					ZERO_EPSILON(0.00001);

	file.setFileName(file_name);
	
	if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_last_error = QStringLiteral("ERR  Cannot open file for writing.");
		return false;
	}
	
	d_graph_positions.clear();

	input_a_data = d_entry_a_widget->Entry_Items();
	input_b_data = d_entry_b_widget->Entry_Items();
	
	// must have at least two entries
	if(input_a_data.size() < 2 ||
	   input_b_data.size() < 2)
	{
		file.close();
		d_last_error = QStringLiteral("ERR  Each dataset must have a minimum of two entries.");
		
		return false;
	}
	
	// data is sorted by length from entry widgets
	min_a_position = input_a_data[0].actual;
	max_a_position = input_a_data[input_a_data.size()-1].actual;
	
	min_b_position = input_b_data[0].actual;
	max_b_position = input_b_data[input_b_data.size()-1].actual;
	
	min_position = min_a_position;
	if(min_b_position < min_position) min_position = min_b_position;
	
	max_position = max_a_position;
	if(max_b_position > max_position) max_position = max_b_position;
	
	for(entry_iter = input_a_data.begin();entry_iter != input_a_data.end();++entry_iter)
	{
		pnt.x = (*entry_iter).actual;
		
		pnt.y = (*entry_iter).actual - (*entry_iter).nominal;
		dev_a_data.push_back(pnt);
		
		pnt.y = (*entry_iter).uc;
		uc_a_data.push_back(pnt);
		
		this->Add_Graph_Position(pnt.x);
	}
	
	for(entry_iter = input_b_data.begin();entry_iter != input_b_data.end();++entry_iter)
	{
		pnt.x = (*entry_iter).actual;
		
		pnt.y = (*entry_iter).actual - (*entry_iter).nominal;
		dev_b_data.push_back(pnt);
		
		pnt.y = (*entry_iter).uc;
		uc_b_data.push_back(pnt);
		
		this->Add_Graph_Position(pnt.x);
		
	}
	
	std::sort(d_graph_positions.begin(), d_graph_positions.end(), position_sort);
	
	// should never fail but just in case
	assert(dev_a_data.size() == uc_a_data.size());
	assert(dev_b_data.size() == uc_b_data.size());
	
	// calculate en score for all a entries if length is common to both
	
	en_count = 0;
	en_mean = 0.0;
	en_max = 0.0;
	init = true;
	
	file.write("Comparison Data\n\n");
	
	//          012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789
	file.write("Position       Dev. A     Dev. B      Uc A       Uc B        En\n");
	file.write("----------------------------------------------------------------------\n");

	for(position_iter = d_graph_positions.begin();position_iter != d_graph_positions.end();++position_iter)
	{
		pnt.x = (*position_iter);
		
		Extrapolation::Get_Value(dev_a_data,pnt.x,&dev_a);
		Extrapolation::Get_Value(dev_b_data,pnt.x,&dev_b);
		
		Extrapolation::Get_Value(uc_a_data,pnt.x,&uc_a);
		Extrapolation::Get_Value(uc_b_data,pnt.x,&uc_b);
		
		text = QString("%1 %2 %3").arg(pnt.x,10,'f',4).arg(dev_a,10,'f',4).arg(dev_b,10,'f',4);
		
		if(pnt.x > (min_a_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_a_position + MAX_EXTRAPOLATION_LENGTH))
		{
			text += QString(" %1").arg(uc_a,10,'f',4);
		}
		else
		{
			text += QStringLiteral("     ------");
		}
		
		if(pnt.x > (min_b_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_b_position + MAX_EXTRAPOLATION_LENGTH))
		{
			text += QString(" %1").arg(uc_b,10,'f',4);
		}
		else
		{
			text += QStringLiteral("     ------");
		}

		// only calculate score is data point is inside range of both data sets
		if(pnt.x > (min_a_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_a_position + MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x > (min_b_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_b_position + MAX_EXTRAPOLATION_LENGTH))
		{
			difference = fabs(dev_a - dev_b);
			uc_ab = sqrt(uc_a * uc_a + uc_b * uc_b);
			
			if(uc_ab > ZERO_EPSILON)
			{
				en = difference / uc_ab;
			}
			else
			{
				en = 1000.0;
			}
			
			text += QString(" %1").arg(en,10,'f',4);
			
			++en_count;
			en_mean += en;
			
			if(en > en_max) en_max = en;
		}
		else
		{
			text += QStringLiteral("     ------");
		}
		
		text += QString('\n');
		
		file.write(text.toLatin1());
	}
	
	if(en_count)
	{
		en_mean /= static_cast<double>(en_count);

		text = QString("\n\nEn Avg: %1\n").arg(en_mean,0,'f',4);
		text += QString("En Max: %1\n").arg(en_max,0,'f',4);
		file.write(text.toLatin1());
	}
	
	file.close();
	
	return true;
}

bool TLengthCompare::Load_Data(
	const QString						&file_name)
{
	TXmlFile							xml_file;
	int									version;
	
	d_current_file_name = file_name;

	try
	{
		xml_file.Set_File_Name(file_name);
		xml_file.Open("Length_Compare_Data",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Data_V1(&xml_file);
				break;
				
			default:
				throw TXmlFileException(QStringLiteral("ERR:  File version not recognized"));
				break;
		}
	}
	catch(TXmlFileException exception)
	{
		d_last_error = QString("ERR  Load_Data: %1").arg(exception.ErrorText()).toLatin1();
		return false;
	}
	
	return true;
}

void TLengthCompare::Load(void)
{
	QString								file_name;
	QString								file_name_filter(QStringLiteral("Length Comparison Files (*.lencomp)"));
	
	if(d_data_changed)
	{
		d_message_box->setText("Data Not Saved");
		d_message_box->setInformativeText("Do you want to continue without saving?");
		d_message_box->setDetailedText(QString());
		d_message_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_message_box->setDefaultButton(QMessageBox::No);
		d_message_box->setIcon(QMessageBox::Warning);
		
		if(QMessageBox::No == d_message_box->exec())
		{
			return;
		}
	}
	
	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Select length comparision file to open"),
											 d_current_file_name,
											 file_name_filter);
	
	if(file_name.length())
	{
		if(!this->Load_Data(file_name))
		{
			d_message_box->setText("Error Loading File");
			d_message_box->setInformativeText("The selected length comparision file cannot be loaded.");
			d_message_box->setDetailedText(d_last_error);
			d_message_box->setStandardButtons(QMessageBox::Ok);
			d_message_box->setDefaultButton(QMessageBox::Ok);
			d_message_box->setIcon(QMessageBox::Critical);
			
			d_message_box->exec();
		}
		
		this->Entry_Data_Changed();
		d_data_changed = false;
	}
}

void TLengthCompare::Save(void)
{
	QString								file_name;
	QString								file_name_filter(QStringLiteral("Length Comparison Files (*.lencomp)"));
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Enter name of length comparision file to create"),
											 d_current_file_name,
											 file_name_filter);
	
	if(file_name.length())
	{
		d_current_file_name = file_name;
		
		if(!this->Save_Data(file_name))
		{
			d_message_box->setText("Error Saving File");
			d_message_box->setInformativeText("The length comparision file cannot be written.");
			d_message_box->setDetailedText(d_last_error);
			d_message_box->setStandardButtons(QMessageBox::Ok);
			d_message_box->setDefaultButton(QMessageBox::Ok);
			d_message_box->setIcon(QMessageBox::Critical);
			
			d_message_box->exec();
		}
		
		d_data_changed = false;
	}
}

void TLengthCompare::Export(void)
{
	QString								file_name;
	QString								default_file_name;
	QString								file_name_filter(QStringLiteral("Text files (*.txt)"));
	
	
	default_file_name = d_current_file_name;
	
	if(default_file_name.endsWith(QStringLiteral(".lencomp"),Qt::CaseInsensitive))
	{
		default_file_name.chop(8);	// remove .lencomp
	}
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Enter name of text file to create"),
											 default_file_name,
											 file_name_filter);
	
	if(file_name.length())
	{
		if(!this->Export_Data(file_name))
		{
			d_message_box->setText("Error Exporting File");
			d_message_box->setInformativeText("The length comparision file cannot be exported.");
			d_message_box->setDetailedText(d_last_error);
			d_message_box->setStandardButtons(QMessageBox::Ok);
			d_message_box->setDefaultButton(QMessageBox::Ok);
			d_message_box->setIcon(QMessageBox::Critical);
			
			d_message_box->exec();
		}
	}
}

void TLengthCompare::Entry_Data_Changed(void)
{
	std::vector<Type::TEntryItem>		input_a_data;
	std::vector<Type::TEntryItem>		input_b_data;
	std::vector<Type::TEntryItem>::const_iterator entry_iter;
	std::vector<TVector2>				dev_a_data;
	std::vector<TVector2>				dev_b_data;
	std::vector<TVector2>				uc_a_data;
	std::vector<TVector2>				uc_b_data;
	TVector2							pnt;
	TVector2							p1,p2,p3,p4;
	Position_Sort						position_sort;
	std::vector<double>::const_iterator	position_iter;
	double								min_a_position(0);
	double								max_a_position(0);
	double								min_b_position(0);
	double								max_b_position(0);
	double								min_en_position(0);
	double								max_en_position(0);
	double								min_position(0);
	double								max_position(0);
	double								dev_a,dev_b;
	double								uc_a,uc_b;
	double								difference;
	double								uc_ab;
	double								en;
	double								uc_radius;
	int									set_a_id;
	int									set_b_id;
	int									set_en_id;
	int									set_id;
	bool								init;
	static const double					MAX_EXTRAPOLATION_LENGTH(10.0);
	static const double					ZERO_EPSILON(0.00001);
	static const double					UC_RADIUS_MULTIPLIER(0.005);
	
	input_a_data = d_entry_a_widget->Entry_Items();
	input_b_data = d_entry_b_widget->Entry_Items();
	
	d_graph_data_widget->DeleteDataSets();
	d_graph_en_widget->DeleteDataSets();
	
	d_graph_positions.clear();
	
	d_data_changed = true;
	
	// must have at least two entries
	if(input_a_data.size() < 2 ||
	   input_b_data.size() < 2)
	{
		d_save_button->setEnabled(false);
		d_export_button->setEnabled(false);
		
		return;
	}
	
	d_save_button->setEnabled(true);
	d_export_button->setEnabled(true);

	
	// data is sorted by length from entry widgets
	min_a_position = input_a_data[0].actual;
	max_a_position = input_a_data[input_a_data.size()-1].actual;
	
	min_b_position = input_b_data[0].actual;
	max_b_position = input_b_data[input_b_data.size()-1].actual;
	
	min_position = min_a_position;
	if(min_b_position < min_position) min_position = min_b_position;
	
	max_position = max_a_position;
	if(max_b_position > max_position) max_position = max_b_position;
	
	for(entry_iter = input_a_data.begin();entry_iter != input_a_data.end();++entry_iter)
	{
		pnt.x = (*entry_iter).actual;
		
		pnt.y = (*entry_iter).actual - (*entry_iter).nominal;
		dev_a_data.push_back(pnt);
		
		pnt.y = (*entry_iter).uc;
		uc_a_data.push_back(pnt);
		
		this->Add_Graph_Position(pnt.x);
	}
	
	for(entry_iter = input_b_data.begin();entry_iter != input_b_data.end();++entry_iter)
	{
		pnt.x = (*entry_iter).actual;
		
		pnt.y = (*entry_iter).actual - (*entry_iter).nominal;
		dev_b_data.push_back(pnt);
		
		pnt.y = (*entry_iter).uc;
		uc_b_data.push_back(pnt);
		
		this->Add_Graph_Position(pnt.x);

	}
	
	std::sort(d_graph_positions.begin(), d_graph_positions.end(), position_sort);
	
	// should never fail but just in case
	assert(dev_a_data.size() == uc_a_data.size());
	assert(dev_b_data.size() == uc_b_data.size());

	// calculate en score for all a entries if length is common to both
	set_a_id = d_graph_data_widget->CreateDataSet(QColor(128,40,40));
	set_b_id = d_graph_data_widget->CreateDataSet(QColor(20,96,20));
	set_en_id = d_graph_en_widget->CreateDataSet(QColor(40,40,128));
	
	uc_radius = (max_position - min_position) * UC_RADIUS_MULTIPLIER;

	init = true;
	
	for(position_iter = d_graph_positions.begin();position_iter != d_graph_positions.end();++position_iter)
	{
		pnt.x = (*position_iter);
		
		Extrapolation::Get_Value(dev_a_data,pnt.x,&dev_a);
		Extrapolation::Get_Value(dev_b_data,pnt.x,&dev_b);
		
		Extrapolation::Get_Value(uc_a_data,pnt.x,&uc_a);
		Extrapolation::Get_Value(uc_b_data,pnt.x,&uc_b);
		
		// only calculate score is data point is inside range of both data sets
		if(pnt.x > (min_a_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_a_position + MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x > (min_b_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_b_position + MAX_EXTRAPOLATION_LENGTH))
		{
			difference = fabs(dev_a - dev_b);
			uc_ab = sqrt(uc_a * uc_a + uc_b * uc_b);
			
			if(uc_ab > ZERO_EPSILON)
			{
				en = difference / uc_ab;
			}
			else
			{
				en = 1000.0;
			}
			
			d_graph_en_widget->AddPoint(set_en_id,TVector2(pnt.x,en));
			
			if(init)
			{
				min_en_position = pnt.x;
				max_en_position = pnt.x;
				
				init = false;
			}
			else
			{
				if(pnt.x < min_en_position) min_en_position = pnt.x;
				if(pnt.x > max_en_position) max_en_position = pnt.x;
			}
 
		}
		
		if(pnt.x > (min_a_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_a_position + MAX_EXTRAPOLATION_LENGTH))
		{
			pnt.y = dev_a;
			d_graph_data_widget->AddPoint(set_a_id,pnt);
			
			p1 = pnt;
			p1.x += (uc_radius/2.0);
			p4 = p1;

			p1.y += uc_a;
			p4.y -= uc_a;
			
			p2 = p1;
			p2.x -= uc_radius;
			
			p3 = p4;
			p3.x = p2.x;
			
			set_id = d_graph_data_widget->CreateDataSet(QColor(212,96,96));
			d_graph_data_widget->AddPoint(set_id,p1);
			d_graph_data_widget->AddPoint(set_id,p2);
			d_graph_data_widget->AddPoint(set_id,p3);
			d_graph_data_widget->AddPoint(set_id,p4);
		}

		if(pnt.x > (min_b_position - MAX_EXTRAPOLATION_LENGTH) &&
		   pnt.x < (max_b_position + MAX_EXTRAPOLATION_LENGTH))
		{
			pnt.y = dev_b;
			d_graph_data_widget->AddPoint(set_b_id,pnt);
			
			p1 = pnt;
			p1.x -= (uc_radius/2.0);
			p4 = p1;
			
			p1.y += uc_b;
			p4.y -= uc_b;
			
			p2 = p1;
			p2.x += uc_radius;
			
			p3 = p4;
			p3.x = p2.x;
			
			set_id = d_graph_data_widget->CreateDataSet(QColor(20,96,20));
			d_graph_data_widget->AddPoint(set_id,p1);
			d_graph_data_widget->AddPoint(set_id,p2);
			d_graph_data_widget->AddPoint(set_id,p3);
			d_graph_data_widget->AddPoint(set_id,p4);
		}
	}
	
	// draw tolerance line for en
	set_en_id = d_graph_en_widget->CreateDataSet(QColor(196,196,196));

	d_graph_en_widget->AddPoint(set_en_id,TVector2(min_position,1.0));
	d_graph_en_widget->AddPoint(set_en_id,TVector2(max_position,1.0));
	
	d_graph_data_widget->UpdateGraph();
	d_graph_en_widget->UpdateGraph();
}

void TLengthCompare::closeEvent(
	QCloseEvent							*event)
{
	// event->ignore();
	
	if(d_data_changed)
	{
		d_message_box->setText("Data Not Saved");
		d_message_box->setInformativeText("Do you want to close without saving?");
		d_message_box->setDetailedText(QString());
		d_message_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_message_box->setDefaultButton(QMessageBox::No);
		d_message_box->setIcon(QMessageBox::Warning);
		
		if(QMessageBox::No == d_message_box->exec())
		{
			event->ignore();
			return;
		}
	}
	
	event->accept();
}


void TLengthCompare::dragEnterEvent(
	QDragEnterEvent 					*drag_enter_event)
{
	if(drag_enter_event->mimeData()->hasFormat("text/uri-list"))
	{
		drag_enter_event->acceptProposedAction();
	}
}

void TLengthCompare::dropEvent(
	QDropEvent 							*drop_event)
{
	QList<QUrl>							urls;
	QString								file_name;
	int									result;
	
	urls = drop_event->mimeData()->urls();
	
	drop_event->acceptProposedAction();
	
	if(urls.size() == 1)
	{
		file_name = (*urls.begin()).toLocalFile();
		
		if(file_name.endsWith(QStringLiteral(".lencomp"),Qt::CaseInsensitive) == true)
		{
			if(d_data_changed)
			{
				d_message_box->setText("Data Not Saved");
				d_message_box->setInformativeText("The length comparision data has been modified.  Do you want to save the changes?");
				d_message_box->setDetailedText(QString());
				d_message_box->setIcon(QMessageBox::Warning);
				d_message_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
				d_message_box->setDefaultButton(QMessageBox::Yes);
				
				result = d_message_box->exec();
				
				switch(result)
				{
					case QMessageBox::Yes:
						this->Save();
						break;
						
					case QMessageBox::No:
						break;
						
					case QMessageBox::Cancel:
					default:
						return;
				}
			}
			
			this->Load_Data(file_name);
			
			this->Entry_Data_Changed();
			d_data_changed = false;
		}
	}
}


void TLengthCompare::Add_Graph_Position(
	const double						&position)
{
	std::vector<double>::const_iterator iter;
	double								separation;
	const double						MIN_SEPARATION(1.0);
	
	for(iter = d_graph_positions.begin();iter != d_graph_positions.end();++iter)
	{
		separation = fabs((*iter) - position);
		
		if(separation < MIN_SEPARATION)
		{
			return;
		}
	}
	
	d_graph_positions.push_back(position);
}


void TLengthCompare::Load_Data_V1(
	TXmlFile							* const xml_file)
{
	QDomElement							e0;
	QDomElement							e1;
	QDomElement							e2;
	std::vector<Type::TEntryItem>		entry_items_a;
	std::vector<Type::TEntryItem>		entry_items_b;
	Type::TEntryItem					entry_item;
	
	assert(xml_file);
	
	d_entry_a_widget->Clear_Entry_Items();
	d_entry_b_widget->Clear_Entry_Items();
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Data_Set_A"));
	e1 = xml_file->Get_Node(&e0,QStringLiteral("Entry"),true);
	
	while(!e1.isNull())
	{
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Nominal"));
		entry_item.nominal = xml_file->Read_Text_Node(&e2).toDouble();
		
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Actual"));
		entry_item.actual = xml_file->Read_Text_Node(&e2).toDouble();
		
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Uc"));
		entry_item.uc = xml_file->Read_Text_Node(&e2).toDouble();
		
		entry_items_a.push_back(entry_item);
		
		e1 = xml_file->Get_Sibling_Node(&e1,QStringLiteral("Entry"),true);
	}
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Data_Set_B"));
	e1 = xml_file->Get_Node(&e0,QStringLiteral("Entry"),true);
	
	while(!e1.isNull())
	{
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Nominal"));
		entry_item.nominal = xml_file->Read_Text_Node(&e2).toDouble();
		
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Actual"));
		entry_item.actual = xml_file->Read_Text_Node(&e2).toDouble();
		
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Uc"));
		entry_item.uc = xml_file->Read_Text_Node(&e2).toDouble();
		
		entry_items_b.push_back(entry_item);
		
		e1 = xml_file->Get_Sibling_Node(&e1,QStringLiteral("Entry"),true);
	}
	
	d_entry_a_widget->Add_Entry_Items(entry_items_a);
	d_entry_b_widget->Add_Entry_Items(entry_items_b);

}

void TLengthCompare::Generate_Test_Data(void)
{
	std::vector<Type::TEntryItem>		input_a_entries;
	std::vector<Type::TEntryItem>		input_b_entries;
	Type::TEntryItem					entry_item;
	
	entry_item.nominal = 12.70001;
	entry_item.actual = 12.70062;
	entry_item.uc = 0.0010;
	input_a_entries.push_back(entry_item);
	
	entry_item.nominal = 179.99893;
	entry_item.actual = 179.9993;
	entry_item.uc = 0.0012;
	input_a_entries.push_back(entry_item);
	
	entry_item.nominal = 359.99881;
	entry_item.actual = 359.99961;
	entry_item.uc = 0.0014;
	input_a_entries.push_back(entry_item);
	
	entry_item.nominal = 539.99787;
	entry_item.actual = 539.99891;
	entry_item.uc = 0.0017;
	input_a_entries.push_back(entry_item);
	
	entry_item.nominal = 719.99833;
	entry_item.actual = 719.99906;
	entry_item.uc = 0.0020;
	input_a_entries.push_back(entry_item);
	
	entry_item.nominal = 899.99806;
	entry_item.actual = 899.99965;
	entry_item.uc = 0.0023;
	input_a_entries.push_back(entry_item);
	

	entry_item.nominal = 9.9999;
	entry_item.actual = 10.00065;
	entry_item.uc = 0.0012;
	input_b_entries.push_back(entry_item);
	
	entry_item.nominal = 209.9929;
	entry_item.actual = 209.9945325;
	entry_item.uc = 0.0016;
	input_b_entries.push_back(entry_item);
	
	entry_item.nominal = 409.9954;
	entry_item.actual = 409.996025;
	entry_item.uc = 0.0019;
	input_b_entries.push_back(entry_item);
	
	entry_item.nominal = 609.9995;
	entry_item.actual = 610.0007775;
	entry_item.uc = 0.0023;
	input_b_entries.push_back(entry_item);
	
	entry_item.nominal = 810.0009;
	entry_item.actual = 810.002225;
	entry_item.uc = 0.0027;
	input_b_entries.push_back(entry_item);
	
	entry_item.nominal = 1010.0003;
	entry_item.actual = 1010.002795;
	entry_item.uc = 0.0031;
	input_b_entries.push_back(entry_item);

	d_entry_a_widget->Add_Entry_Items(input_a_entries);
	d_entry_b_widget->Add_Entry_Items(input_b_entries);
}

